# sqltail

from sqltail.monitor import SQLTail
from sqltail.db import Database, DatabaseException, DatabaseConnectionFailed, DatabaseNotFound

__version__='1.0.3'
__license__='Copyright 2021 Reliance Systems, Inc. - MIT license - https://github.com/rstms/sqltail'
